import json
import logging
import pathlib
from datetime import datetime
from typing import List, Optional, Union

import click
import yaml
from cspark.sdk import BaseUrl, LoggerOptions, SparkError
from cspark.sdk._validators import Validators
from cspark.sdk._version import sdk_logger
from InquirerPy import inquirer
from InquirerPy.validator import PathValidator
from rich.console import Console
from rich.text import Text

from .._utils import DATE_FORMAT, Profile, add_or_update_profile, delete_profile, get_active_profile, load_profiles
from ._api import AliasedGroup, parse_pairs

_SPARK_ENVS = ['uat.us', 'uat.eu', 'uat.jp', 'uat.ca', 'uat.au', 'us', 'ca', 'eu', 'jp', 'au', 'sit', 'dev', 'test']


@click.group(
    name='config',
    help='Manage configuration profiles',
    invoke_without_command=True,
    cls=AliasedGroup,
    aliases={'ls': 'list', 'rm': 'remove', 'del': 'remove'},
)
@click.option(
    '-p',
    '--profile',
    type=bool,
    is_flag=True,
    default=False,
    help='Add a configuration profile to connect to Spark',
)
@click.pass_context
def config_cmd(ctx: click.Context, profile: bool):
    if profile:
        click.echo('🚀 Setting up a new configuration profile')
        build_profile()
        return

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@click.command(name='init', help='Start a new configuration profile')
@click.option(
    '-f',
    '--file',
    type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, readable=True),
    metavar='<CONFIG PATH>',
    help='Path to the configuration file',
)
@click.option('--config', type=str, metavar='<JSON>', help='Configuration profile in JSON format')
def init_cmd(file: Union[pathlib.Path, str, None], config: Optional[str]):
    def msg(name):
        return f'[green]✓[/green] Profile [magenta]{name}[/magenta] created and set as active'

    console = Console()
    if file:
        profile = load_config_file(pathlib.Path(file))
        add_or_update_profile(profile)
        console.print(msg(profile.name))
    elif config:
        try:
            profile = Profile.from_dict(json.loads(config))
            add_or_update_profile(profile)
            console.print(msg(profile.name))
        except json.JSONDecodeError as err:
            Console().print('[red]✗[/red] Invalid JSON configuration')
            raise click.exceptions.Exit(1) from err
    else:
        # FIXME: improve welcome message.
        console.print('Welcome to Coherent Spark CLI!\n\n🚀 Build a configuration profile to connect to Spark')
        build_profile()


def load_config_file(file: pathlib.Path) -> Profile:
    if not file.exists():
        Console().print(f'[red]✗[/red] File not found: {file}')
        raise click.exceptions.Exit(1)

    try:
        if file.suffix.lower() in ['.yml', '.yaml']:
            with file.open() as f:
                config = yaml.safe_load(f)
        else:
            with file.open() as f:
                config = json.load(f)
        return Profile.from_dict(config)
    except (yaml.YAMLError, json.JSONDecodeError) as err:
        Console().print(f'[red]✗[/red] Invalid configuration file: {file}')
        raise click.exceptions.Exit(1) from err


def build_profile():
    profiles = load_profiles(is_init=True)

    profile_name = _collect_profile_name([p.name for p in profiles])
    base_url = _collect_base_url()
    auth = _collect_auth()
    logger = _collect_logger_options()
    timeout, max_retries, retry_interval = _collect_advanced_settings()

    now = datetime.now().isoformat()
    oauth = auth.get('oauth', {})
    profile = Profile(
        name=profile_name,
        is_active=True,
        created_at=now,
        updated_at=now,
        logger=logger,
        base_url=base_url,
        api_key=auth.get('api_key'),
        token=auth.get('token'),
        client_id=oauth.get('client_id') if isinstance(oauth, dict) else None,
        client_secret=oauth.get('client_secret') if isinstance(oauth, dict) else None,
        oauth_path=oauth if isinstance(oauth, str) else None,
        timeout=timeout,
        max_retries=max_retries,
        retry_interval=retry_interval,
    )

    add_or_update_profile(profile)
    Console().print(f'[green]✓[/green] Profile [magenta]{profile_name}[/magenta] created and set as active')


def _collect_profile_name(old_names: list[str]) -> str:
    return inquirer.text(  # type: ignore
        message='Enter profile name:',
        validate=lambda entry: len(entry) > 0 and entry not in old_names,
        filter=lambda entry: entry.strip(),
        invalid_message='Profile name is required and must be unique!',
    ).execute()


def _collect_base_url() -> str:
    options = {'full_url': 'Base URL', 'env_tenant': 'Environment and Tenant'}
    is_empty_str = Validators.empty_str().is_valid

    option = inquirer.select(  # type: ignore
        message='Define Spark connection settings using',
        choices=[*options.values()],
    ).execute()

    if option == options['full_url']:
        url = inquirer.text(  # type: ignore
            message='Enter base URL:',
            validate=Validators.base_url().is_valid,
            invalid_message='Invalid Spark URL',
            completer={'https://excel': None, 'coherent.global': None},
        ).execute()

        try:
            base_url = BaseUrl.of(url=url).full
        except SparkError:
            tenant = inquirer.text(  # type: ignore
                message='Tenant name is missing',
                validate=is_empty_str,
                invalid_message='Tenant name is required',
            ).execute()
            base_url = BaseUrl.of(url=url, tenant=tenant).full
    else:
        env = inquirer.text(  # type: ignore
            message='Enter environment:',
            validate=is_empty_str,
            invalid_message='Environment is required',
            completer={e: None for e in _SPARK_ENVS},
        ).execute()
        tenant = inquirer.text(  # type: ignore
            message='Enter tenant name:',
            validate=is_empty_str,
            invalid_message='Tenant name is required',
        ).execute()
        base_url = BaseUrl.of(env=env, tenant=tenant).full

    confirm_url = inquirer.confirm(  # type: ignore
        message=f'Is this base URL correct: {base_url}?', default=True
    ).execute()

    return base_url if confirm_url else _collect_base_url()


def _collect_auth():
    options = {'api_key': 'API key', 'bearer_token': 'Bearer token', 'oauth2': 'OAuth2 CCG'}
    scheme = inquirer.select(  # type: ignore
        message='Select authorization scheme:',
        choices=[*options.values()],
    ).execute()

    auth = {}
    is_empty_str = Validators.empty_str().is_valid

    if scheme == options['api_key']:
        api_key = inquirer.secret(message='Enter API key:', validate=is_empty_str).execute()  # type: ignore
        auth.update({'api_key': api_key})
    elif scheme == options['bearer_token']:
        token = inquirer.secret(message='Enter Bearer token:', validate=is_empty_str).execute()  # type: ignore
        auth.update({'token': token})
    else:
        oauth_method = inquirer.select(  # type: ignore
            message='Select OAuth2 CCG entry:',
            choices=['Path', 'Client ID and Secret'],
        ).execute()
        if oauth_method == 'Path':
            path = inquirer.filepath(  # type: ignore
                message='Enter path to OAuth2 CCG file:',
                validate=PathValidator(message='Invalid file path', is_file=True),
            ).execute()
            auth.update({'oauth': path})
        else:
            id = inquirer.secret(message='Enter client ID:', validate=is_empty_str).execute()  # type: ignore
            secret = inquirer.secret(message='Enter client secret:', validate=is_empty_str).execute()  # type: ignore
            auth.update({'oauth': {'client_id': id, 'client_secret': secret}})
    return auth


def _collect_logger_options() -> Union[bool, LoggerOptions]:
    enabled = inquirer.confirm(message='Enable logger?', default=False).execute()  # type: ignore

    if not enabled:
        return False

    colorful = inquirer.confirm(message='Should logs be colorful?', default=True).execute()  # type: ignore
    timestamp = inquirer.confirm(message='Should logs have timestamps?', default=True).execute()  # type: ignore
    context = inquirer.text(message='Enter logger context:', default=sdk_logger).execute()  # type: ignore
    level = inquirer.select(  # type: ignore
        message='Select log level:',
        default='DEBUG',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        filter=lambda choice: getattr(logging, choice),
    ).execute()

    return LoggerOptions(context=context, disabled=False, level=level, colorful=colorful, timestamp=timestamp)


def _collect_advanced_settings():
    should_configure = inquirer.confirm(  # type: ignore
        message='Do you want to configure advanced settings?', default=False
    ).execute()

    if not should_configure:
        return None, None, None

    timeout = inquirer.text(  # type: ignore
        message='Timeout in ms:',
        default='60000',
        validate=lambda entry: entry.isdigit(),
        invalid_message='Timeout must be a number!',
    ).execute()

    max_retries = inquirer.text(  # type: ignore
        message='Maximum retries:',
        default='2',
        validate=lambda entry: entry.isdigit(),
        invalid_message='Maximum retries must be a number!',
    ).execute()

    retry_interval = inquirer.text(  # type: ignore
        message='Retry interval in seconds:',
        default='1',
        validate=lambda entry: entry.isdigit(),
        invalid_message='Retry interval must be a number!',
    ).execute()

    return float(timeout), int(max_retries), float(retry_interval)


class ConfigSwitchCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='switch',
            help='Switch between different configuration profiles',
            options_metavar='',
            add_help_option=False,
            callback=self.switch,
        )

    def switch(self):
        profiles = load_profiles()
        selected = inquirer.select(  # type: ignore
            message='Select profile to switch to:',
            choices=[p.name for p in profiles],
        ).execute()

        for p in profiles:
            if p.name == selected:
                p.is_active = True
                add_or_update_profile(p)
                break

        Console().print(f'🔄 Profile switched to [magenta]{selected}[/magenta]!')


class ConfigSetCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='set',
            help='Set configuration values of the active profile',
            options_metavar='',
            add_help_option=False,
            callback=self.set,
            params=[click.Argument(['values'], type=str, nargs=-1)],
        )

    def set(self, values: list[str]):
        updates = parse_pairs(values, sep='=', infer_type=True)
        if len(updates) == 0:
            raise click.UsageError('no configuration values to set; expecting <key>=<value> format')

        profile = get_active_profile()
        all_keys = profile.__dict__.keys()
        for key, val in updates.items():
            if key in all_keys:
                profile.__dict__[key] = val

        profile.updated_at = datetime.now().isoformat()
        add_or_update_profile(profile)
        Console().print(f'Profile [magenta]{profile.name}[/magenta] updated!')


class ConfigGetCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='get',
            help='Read configuration values of the active profile',
            callback=self.get,
            options_metavar='',
            add_help_option=False,
            params=[click.Argument(['key'], type=str)],
        )

    def get(self, key: str):
        profile = get_active_profile()

        key = key.lower()
        if key in ['tenant', 'env', 'environment', 'region', 'url']:
            self._get_url_keys(key, profile)
        elif key in ['auth', 'oauth']:
            self._get_auth_keys(key, profile)
        elif key in profile.__dict__.keys():
            click.echo(profile.__dict__[key])
        else:
            raise click.UsageError(f'"{key}" is not valid key')

    def _get_url_keys(self, key: str, profile: Profile):
        url = profile.extract_url()
        if key == 'tenant':
            click.echo(url.tenant)
        elif key == 'url':
            click.echo(url.value)
        else:
            click.echo(url.env)

    def _get_auth_keys(self, key: str, profile: Profile):
        auth = profile.mask_auth(show=True)
        if key == 'auth':
            for k, v in auth.items():
                click.echo(f'{k}: {v}')
        else:
            config = profile.to_config()
            if 'oauth' in config:
                oauth = config['oauth']
                if isinstance(oauth, dict):
                    for k, v in oauth.items():
                        click.echo(f'{k}: {v}')
                else:
                    click.echo(oauth)
            else:
                click.echo(config.get('token'))


class ConfigListCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='list',
            help='List configuration profiles',
            options_metavar='',
            add_help_option=False,
            callback=self.list,
            params=[
                click.Option(
                    ['-a', '--all'],
                    is_flag=True,
                    default=False,
                    type=bool,
                    help='Display details of all existing profiles',
                ),
                click.Option(
                    ['-v', '--verbose'],
                    is_flag=True,
                    default=False,
                    type=bool,
                    help='Display additional details of a profile',
                ),
            ],
        )

    def list(self, all: bool, verbose: bool):
        profiles = load_profiles()

        if not all and not verbose:
            self._show_profiles(profiles)
        elif all:
            for idx, profile in enumerate(profiles):
                self._display_profile(profile, nl=idx < len(profiles) - 1, verbose=verbose)
        else:
            for profile in profiles:
                if profile.is_active:
                    self._display_profile(profile, nl=False, verbose=verbose)
                    break

    def _show_profiles(self, profiles: List[Profile]):
        console = Console()
        for p in profiles:
            if p.is_active:
                console.print(f'* [magenta]{p.name}[/magenta]')
            else:
                console.print(f'  {p.name}')

    def _display_profile(self, profile: Profile, verbose: bool = False, nl: bool = False):
        url = profile.extract_url()

        console = Console()
        console.print(Text(f'{profile.name}', style=f"{'magenta' if profile.is_active else ''}"))
        console.print(f'  - base URL: {url.value}')
        console.print(f'  - tenant  : {url.tenant}')
        for key, value in profile.mask_auth().items():
            console.print(f'  - {key.ljust(8)}: {value}')

        if verbose:
            updated_at = datetime.fromisoformat(profile.updated_at).strftime(DATE_FORMAT)
            timeout = f'{profile.timeout / 1000.0} second(s)' if profile.timeout else 'None'
            retry_interval = f'{profile.retry_interval} second(s)' if profile.retry_interval else 'None'
            max_retries = f'up to {profile.max_retries} times' if profile.max_retries else 'None'
            console.print(f'  [cyan]other settings[/cyan]')
            console.print(f'    - logger     : [yellow]{"enabled" if profile.logger else "disabled"}[/yellow]')
            console.print(f'    - timeout    : {timeout}')
            console.print(f'    - interval   : {retry_interval}')
            console.print(f'    - max retries: {max_retries}')
            console.print(f'    - updated at : {updated_at}')

        if nl:
            console.print()


class ConfigRemoveCommand(click.Command):
    def __init__(self):
        super().__init__(
            name='remove',
            help='Remove a configuration profile',
            options_metavar='',
            add_help_option=False,
            callback=self.remove,
            params=[click.Option(['-f', '--force'], is_flag=True, default=False, help='Force removal of a profile')],
        )

    def remove(self, force: bool):
        profiles = load_profiles()
        console = Console()

        if len(profiles) == 1:
            if force:
                delete_profile(profiles[0])
                console.print('[red]✗[/red] Active profile removed')
            else:
                console.print(f'[yellow]WARNING: only one profile exists =>[/yellow] {profiles[0].name}')
                console.print('Use [green]--force[/green] to remove the active profile!')
                return
        else:
            selected = inquirer.select(  # type: ignore
                message='Select profile to remove:',
                choices=[p.name for p in profiles] + ['<Cancel>'],
            ).execute()

            for p in profiles:
                if p.name != selected:
                    continue

                if p.is_active and not force:
                    console.print(
                        f'[magenta]{selected}[/magenta] is the active profile; use [green]--force[/green] '
                        'to remove it or switch to another profile before removing it.\n'
                        'Use [green]cspark config switch[/green] to switch profiles.\n'
                        '[yellow]WARNING: Removing the active profile will fallback to the next available profile.[/yellow]'
                    )
                else:
                    delete_profile(p)
                    Console().print(f'[red]✗[/red] Profile [magenta]{selected}[/magenta] removed')
                break


config_cmd.add_command(ConfigSwitchCommand())
config_cmd.add_command(ConfigSetCommand())
config_cmd.add_command(ConfigGetCommand())
config_cmd.add_command(ConfigListCommand())
config_cmd.add_command(ConfigRemoveCommand())



from typing import Any, Mapping, Optional, Union

import cspark.sdk.resources as API
from httpx import Client as HttpClient

from ._auth import Authorization
from ._config import BaseUrl, Config, HealthUrl
from ._logger import LoggerOptions
from .resources._base import download as download_file

__all__ = ['Client']


class Client:
    """
    The main entry point for the Coherent Spark SDK client.

    This class provides access to all the resources available in the SDK, including
    the `folders`, and `services` APIs, as well as the `impex` and `wasm` utilities.
    Do note that this client is synchronous and blocking. For an asynchronous client,
    use the `AsyncClient` class.

    Visit the main documentation page for more details on how to use the SDK.
    See https://github.com/Coherent-Partners/spark-python-sdk/blob/main/docs
    to learn more about the SDK and its capabilities.
    """

    def __init__(
        self,
        *,
        base_url: Union[None, str, BaseUrl] = None,
        tenant: Optional[str] = None,
        env: Optional[str] = None,
        oauth: Union[None, Mapping[str, str], str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[float] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        http_client: Optional[HttpClient] = None,
        logger: Union[bool, Mapping[str, Any], LoggerOptions] = True,
    ) -> None:
        self.config = Config(
            base_url=base_url,
            tenant=tenant,
            env=env,
            oauth=oauth,
            api_key=api_key,
            token=token,
            timeout=timeout,
            max_retries=max_retries,
            retry_interval=retry_interval,
            http_client=http_client,
            logger=logger,
        )

    @property
    def health(self) -> API.Health:
        """The resource to manage health checks."""
        return API.Health(self.config)

    @property
    def folders(self) -> API.Folders:
        """The resource to manage Folders API."""
        return API.Folders(self.config)

    @property
    def services(self) -> API.Services:
        """The resource to manage Services API."""
        return API.Services(self.config)

    @property
    def transforms(self) -> API.Transforms:
        """The resource to manage Transforms API."""
        return API.Transforms(self.config)

    @property
    def batches(self) -> API.Batches:
        """The resource to manage asynchronous batch processing."""
        return API.Batches(self.config)

    @property
    def logs(self) -> API.History:
        """The resource to manage service execution logs."""
        return API.History(self.config)

    @property
    def files(self) -> API.Files:
        """The resource to manage files."""
        return API.Files(self.config)

    @property
    def wasm(self) -> API.Wasm:
        """The resource to manage a service's WebAssembly module."""
        return API.Wasm(self.config)

    @property
    def impex(self) -> API.ImpEx:
        """The resource to import and export Spark services."""
        return API.ImpEx.only(self.config)

    @staticmethod
    def health_check(base_url: Union[str, BaseUrl], token: str = 'open', **options):
        """Checks the health status of the Coherent Spark environment."""
        config = Config(base_url=HealthUrl.when(base_url), token=token, **options)
        with API.Health(config) as health:
            return health.check()

    @staticmethod
    def download(url: str, auth: Optional[Authorization] = None) -> bytes:
        """Downloads a file from the given URL."""
        return download_file(url, headers=auth.as_header if auth else {}).buffer


class AsyncClient:
    pass



import json
import time
from typing import BinaryIO, List, Mapping, Optional, Union

from .._config import Config
from .._constants import SPARK_SDK
from .._errors import RetryTimeoutError, SparkError
from .._utils import get_retry_timeout
from ._base import ApiResource, Uri, UriParams

__all__ = ['ImpEx', 'Export', 'Import', 'Migration', 'Wasm', 'Files']


class ImpEx:
    def __init__(self, config: Config):
        self.config = config

    @staticmethod
    def only(config: Config):
        return ImpEx(config)

    @staticmethod
    def migration(exports: Config, imports: Config):
        return Migration(exports=exports, imports=imports)

    @property
    def exports(self):
        return Export(self.config)

    @property
    def imports(self):
        return Import(self.config)

    def exp(
        self,
        *,
        folders: Optional[List[str]] = None,
        services: Optional[List[str]] = None,
        version_ids: Optional[List[str]] = None,
        file_filter: Optional[str] = None,
        version_filter: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        with self.exports as exporter:
            response = exporter.initiate(
                folders=folders,
                services=services,
                version_ids=version_ids,
                file_filter=file_filter,
                version_filter=version_filter,
                source_system=source_system,
                correlation_id=correlation_id,
            )
            status = exporter.get_status(
                job_id=response.data['id'],  # type: ignore
                max_retries=max_retries,
                retry_interval=retry_interval,
            )

            files = isinstance(status.data, dict) and status.data.get('outputs', {}).get('files', []) or []
            if len(files) == 0:
                error = SparkError.sdk('export job failed to produce any files', status)
                exporter.logger.error(error.message)
                raise error

            return exporter.download([f['file'] for f in files])

    def imp(
        self,
        destination: Union[str, List[str], Mapping[str, str], List[Mapping[str, str]]],
        file: BinaryIO,
        *,
        if_present: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        with self.imports as importer:
            response = importer.initiate(
                destination=destination,
                file=file,
                if_present=if_present,
                source_system=source_system,
                correlation_id=correlation_id,
            )
            status = importer.get_status(
                job_id=response.data['id'],  # type: ignore
                max_retries=max_retries,
                retry_interval=retry_interval,
            )

            services = isinstance(status.data, dict) and status.data.get('outputs', {}).get('services', []) or []
            if isinstance(status.data, dict) and status.data.get('errors'):
                error = SparkError.sdk('import job failed with errors', status)
                importer.logger.error(error.message)
                raise error
            elif len(services) == 0:
                importer.logger.warning('import job completed without any services')
            else:
                importer.logger.info(f'{len(services)} service(s) imported')

            return status

    # aliases for export and import to avoid breaking changes.
    export = exp
    import_ = imp


class Export(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.full, 'version': 'api/v4'}

    def initiate(
        self,
        *,
        folders: Optional[List[str]] = None,
        services: Optional[List[str]] = None,
        version_ids: Optional[List[str]] = None,
        file_filter: Optional[str] = None,
        version_filter: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
    ):
        metadata = {
            'file_filter': file_filter or 'migrate',
            'version_filter': version_filter or 'latest',
            'source_system': source_system or SPARK_SDK,
            'correlation_id': correlation_id,
        }
        inputs = {
            k: v
            for k, v in {
                'folders': folders,
                'services': services,
                'version_ids': version_ids,
            }.items()
            if v
        }

        if len(inputs) == 0:
            error = SparkError.sdk('at least one of folders, services, or version_ids must be provided')
            self.logger.error(error.message)
            raise error

        url = Uri.of(None, endpoint='export', **self._base_uri)
        response = self.request(url, method='POST', body={'inputs': inputs, **metadata})
        if isinstance(response.data, dict):
            self.logger.info(f'export job created <{response.data["id"]}>')
        return response

    def get_status(
        self,
        *,
        job_id: Optional[str] = None,
        url: Optional[str] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        if job_id is None and url is None:
            error = SparkError.sdk('either job_id or url must be provided')
            self.logger.error(error.message)
            raise error

        max_retries = max_retries or self.config.max_retries
        retry_interval = retry_interval or self.config.retry_interval
        status_url = Uri.of(None, endpoint=f'export/{job_id}/status', **self._base_uri) if job_id else url

        retries = 0
        while retries < max_retries:
            response = self.request(status_url)  # type: ignore
            if isinstance(response.data, dict) and response.data.get('status') in ['completed', 'closed']:
                self.logger.info(f'export job <{job_id}> completed')
                return response

            retries += 1
            self.logger.info(f'waiting for export job to complete (attempt {retries} of {max_retries})')
            delay = get_retry_timeout(retries, retry_interval)
            time.sleep(delay)

        err_msg = f'export job status timed out after {retries} attempts'
        self.logger.error(err_msg)
        raise RetryTimeoutError(err_msg, retries=retries, interval=retry_interval)

    def cancel(self, job_id: str):
        url = Uri.of(None, endpoint=f'export/{job_id}', **self._base_uri)
        response = self.request(url, method='PATCH', body={'export_status': 'cancelled'})
        if isinstance(response.data, dict):
            self.logger.info(f'export job <${response.data["id"]}> has been cancelled')
        return response

    def download(self, urls: Union[str, List[str]]):
        downloads = []

        urls = urls if isinstance(urls, list) else [urls]
        for url in urls:
            if not url:
                self.logger.warning('skipping empty download url')
                continue
            try:
                downloads.append(self.request(url))
            except Exception as cause:
                self.logger.warning(f'failed to download file <{url}>', cause)

        return downloads


class Import(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)
        self._base_uri = {'base_url': self.config.base_url.full, 'version': 'api/v4'}

    def initiate(
        self,
        destination: Union[str, List[str], Mapping[str, str], List[Mapping[str, str]]],
        file: BinaryIO,
        *,
        if_present: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
    ):
        files = {'file': ('package.zip', file, 'application/zip')}
        metadata = {
            'inputs': {'services_modify': _build_service_mappings(destination)},
            'services_existing': if_present or 'add_version',
            'source_system': source_system or SPARK_SDK,
            'correlation_id': correlation_id,
        }

        url = Uri.of(None, endpoint='import', **self._base_uri)
        response = self.request(url, method='POST', form={'importRequestEntity': json.dumps(metadata)}, files=files)
        if isinstance(response.data, dict):
            self.logger.info(f'import job created <{response.data["id"]}>')
        return response

    def get_status(
        self,
        *,
        job_id: Optional[str] = None,
        url: Optional[str] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
    ):
        if job_id is None and url is None:
            error = SparkError.sdk('either job_id or url must be provided')
            self.logger.error(error.message)
            raise error

        max_retries = max_retries or self.config.max_retries
        retry_interval = retry_interval or self.config.retry_interval
        status_url = Uri.of(None, endpoint=f'import/{job_id}/status', **self._base_uri) if job_id else url

        retries = 0
        while retries < max_retries:
            response = self.request(status_url)  # type: ignore
            if isinstance(response.data, dict) and response.data.get('status') in ['completed', 'closed']:
                self.logger.info(f'import job <{job_id}> completed')
                return response

            retries += 1
            self.logger.info(f'waiting for import job to complete (attempt {retries} of {max_retries})')
            delay = get_retry_timeout(retries, retry_interval)
            time.sleep(delay)

        err_msg = f'import job status timed out after {retries} attempts'
        self.logger.error(err_msg)
        raise RetryTimeoutError(err_msg, retries=retries, interval=retry_interval)


class Migration:
    def __init__(self, *, exports: Config, imports: Config):
        self.configs = {'exports': exports, 'imports': imports}

    @property
    def exports(self):
        return Export(self.configs['exports'])

    @property
    def imports(self):
        return Import(self.configs['imports'])

    def migrate(self):
        raise NotImplementedError('not implemented yet')


class Wasm(ApiResource):
    def download(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        folder: Optional[str] = None,
        service: Optional[str] = None,
        service_id: Optional[str] = None,
        version_id: Optional[str] = None,
        public: Optional[bool] = False,
    ):
        params = (
            UriParams(folder, service, service_id, version_id=version_id, public=public)
            if uri is None
            else Uri.to_params(uri)
        )
        service_uri = Uri.validate(params).encode()
        endpoint = f'getnodegenzipbyId/{service_uri}'
        resource = 'nodegen' + ('/public' if params.public else '')
        url = Uri.partial(resource, base_url=self.config.base_url.full, endpoint=endpoint)

        return self.request(url)


class Files(ApiResource):
    def download(self, url: str):
        return self.request(url)


def _build_service_mappings(
    uri: Union[str, List[str], Mapping[str, str], List[Mapping[str, str]]], upgrade_type: str = 'minor'
) -> List[Mapping[str, str]]:
    if isinstance(uri, str):
        return [
            {
                'service_uri_source': uri,
                'service_uri_destination': uri,
                'update_version_type': upgrade_type,
            }
        ]

    if isinstance(uri, list):
        uris: List[Mapping[str, str]] = []
        for u in uri:
            mappings = _build_service_mappings(u, upgrade_type)
            uris.extend(mappings)
        return uris

    if isinstance(uri, dict):
        source = uri.get('source', '')
        return [
            {
                'service_uri_source': source,
                'service_uri_destination': uri.get('target', source),
                'update_version_type': uri.get('upgrade', upgrade_type),
            }
        ]

    raise SparkError.sdk('invalid import service uri', uri)



try:
    from ._main import __version__, main  # noqa: F401
except ImportError:

    def main() -> None:
        import sys

        print(  # noqa: T201
            'The cspark command-line client could not run because the required '
            'dependencies were not installed.\nMake sure you have installed '
            f"everything with: \x1b[32mpip install 'cspark[cli]'\x1b[39m"
        )
        sys.exit(1)
